/*
 *
 * Copyright(C) OASIS(R) 2009,2010. All Rights Reserved.
 * OASIS trademark, IPR and other policies apply.    
 *     
 */
package org.oasisopen.sca.test;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import org.oasisopen.sca.annotation.*;
import org.oasisopen.sca.ComponentContext;
import org.oasisopen.sca.RequestContext;
import org.oasisopen.sca.ServiceReference;
import org.oasisopen.sca.test.Service1;

/**
 * Java component implementation with
 * 1 service with interface Service1
 * 2 references
 * - singleRef (1..1) with interface Service1
 * - callback3Ref (1..1) with interface Service3WithCallback 
 * 
 * This implementation performs invocations of the RequestContext getServiceName() method of the API and
 * checks that the results conform to a series of specification requirements
 */
@Service({Service1.class, Service2.class, Service3WithCallback.class})
public class Service1RCgetServiceNameImpl implements Service1, Service2, Service3WithCallback, Service3Callback {
	
	// Marked "volatile" due to sharing between multiple threads - synchonization not an issue since the test is
	// organized so that only one thread can possibly be updating these fields at any one time
	private volatile RequestContext requestContext;
	private volatile String responseAccumulator = "";
	
	// Lock and associated condition used to signal the original thread that all of the parallel invocations are complete
	private ReentrantLock completionLock = new ReentrantLock();
	private Condition     completionCondition = completionLock.newCondition();
	
	@Property
	public String serviceName = "service1";
	@Reference
	public Service1 singleRef;
	@Reference
	public Service3WithCallback callback3Ref;

	public Service1RCgetServiceNameImpl() {
		super();
	}
	
	/**
	 * Setter for the RequestContext used by this implementation
	 * @param theContext - the RequestContext
	 */
	@Context
	public void setContext( RequestContext theContext ) {
		requestContext = theContext;
	} // end method setContext

	public String operation1(String input) {
		responseAccumulator += serviceName + " operation1 invoked RequestContext";
		
		if( requestContext == null ) return responseAccumulator + " injected requestContext null";
		
		// JCA-TA-9039 - returns the Name of the service invoked
		try {
		    String serviceName = requestContext.getServiceName();
		    if ( serviceName != null ) {
			    responseAccumulator += " expected non-null serviceName from getServiceName";
			    if( serviceName.equals("Service1")) {
			    	responseAccumulator += " expected value for returned serviceName";
			    } else {
			    	responseAccumulator += " unexpected value for returned serviceName: " + serviceName;
			    }
		    } else {
			    responseAccumulator += " unexpected returned serviceName null from getServiceName";
		    }
		} catch ( IllegalArgumentException e ) {
			responseAccumulator += " unexpected IllegalArgumentException received from getServiceName";	
		} catch ( Exception e ) {
			responseAccumulator += " unexpected other exception received from getServiceName: " + e.toString();
		} // end try
	
		// JCA-TA-9042 - getCallbackReference returns null when called from a thread processing an operation of 
		// a non-bidirectional service
		try {
		    ServiceReference<?> callbackReference = requestContext.getCallbackReference();
		    if ( callbackReference != null ) {
			    responseAccumulator += " unexpected non-null ServiceReference from getCallbackReference";
		    } else {
			    responseAccumulator += " expected null ServiceReference from getCallbackReference";
		    }
		} catch ( IllegalArgumentException e ) {
			responseAccumulator += " unexpected IllegalArgumentException received from getCallbackReference";	
		} catch ( Exception e ) {
			responseAccumulator += " unexpected other exception received from getCallbackReference: " + e.toString();
		} // end try
		
		// JCA-TA-9045 - getCallback returns null when called from a thread processing an operation of 
		// a non-bidirectional service
		try {
		    Object callbackReference = requestContext.getCallback();
		    if ( callbackReference != null ) {
			    responseAccumulator += " unexpected non-null reference proxy from getCallback";
		    } else {
			    responseAccumulator += " expected null reference proxy from getCallback";
		    }
		} catch ( IllegalArgumentException e ) {
			responseAccumulator += " unexpected IllegalArgumentException received from getCallback";	
		} catch ( Exception e ) {
			responseAccumulator += " unexpected other exception received from getCallback: " + e.toString();
		} // end try
		
		// Invoke the singleRef reference which will cause an invocation of the Service3WithCallback service...
		responseAccumulator += singleRef.operation1("start callback");
		
		// Invoke the callback3Ref which invokes a Service3WithCallback service and results in the Service3Callback method of
		// this class to be invoked as a callback
		responseAccumulator += callback3Ref.operation3("invoke callback service");
		
		runSeparateThreadTests();

		return responseAccumulator;
	} // end operation1

	public int operation2(int input) {
		// Dummy implementation of the method of Service2 service...
		return 0;
	}
	
	/**
	 * Method used to run a set of tests on a separate thread, to deal with those test assertions that relate to the
	 * invocation of the APIs from a thread that is not processing an SCA service or callback operation
	 */
	private void runSeparateThreadTests() {
		
		int invocationCount = 2;	// # of threads to use
		long maxWaitTime = 5000;	// Max wait time for completion = 5sec		
		
		// Run the tests using a ThreadPoolExecutor
		ThreadPoolExecutor theExecutor = new ThreadPoolExecutor( 	invocationCount,
													                invocationCount,
													                maxWaitTime,
													                TimeUnit.MILLISECONDS,
													                new ArrayBlockingQueue<Runnable>( invocationCount ) );
		completionLock.lock();
		try {
			// Perform the invocations on separate thread...
		    theExecutor.execute( new separateThreadInvoker() );
		    // Wait for completion or for timeout
		    completionCondition.await( maxWaitTime, TimeUnit.MILLISECONDS);
		} catch (InterruptedException e ) {
			// nothing to do
		} finally {
		    completionLock.unlock();
		} // end try
	}

	/**
	 * An inner class which acts as a runnable task for invoking APIs on threads that are not processing
	 * either a service operation or a callback operation
	 */
	private class separateThreadInvoker implements Runnable {
				
		public separateThreadInvoker() {
			super();
		} // end constructor

		public void run() {
			
			// JCA-TA-9040 - getServiceName returns null when called from a thread not processing service or callback operation
			try {
			    String serviceName = requestContext.getServiceName();
			    if ( serviceName != null ) {
				    responseAccumulator += " unexpected non-null serviceName from getServiceName";
			    } else {
				    responseAccumulator += " expected null from getServiceName";
			    }
			} catch ( IllegalArgumentException e ) {
				responseAccumulator += " unexpected IllegalArgumentException received from getServiceName";	
			} catch ( Exception e ) {
				responseAccumulator += " unexpected other exception received from getServiceName: " + e.toString();
			} // end try
			
			// JCA-TA-9043 - getCallbackReference returns null when called from a thread not processing an operation of 
			// a service
			try {
			    ServiceReference<?> callbackReference = requestContext.getCallbackReference();
			    if ( callbackReference != null ) {
				    responseAccumulator += " unexpected non-null ServiceReference from getCallbackReference";
			    } else {
				    responseAccumulator += " expected null ServiceReference from getCallbackReference";
			    }
			} catch ( IllegalArgumentException e ) {
				responseAccumulator += " unexpected IllegalArgumentException received from getCallbackReference";	
			} catch ( Exception e ) {
				responseAccumulator += " unexpected other exception received from getCallbackReference: " + e.toString();
			} // end try			

			// JCA-TA-9046 - getCallback returns null when called from a thread not processing an operation of 
			// a service
			try {
				Object callbackReference = requestContext.getCallback();
			    if ( callbackReference != null ) {
				    responseAccumulator += " unexpected non-null reference proxy from getCallback";
			    } else {
				    responseAccumulator += " expected null reference proxy from getCallback";
			    }
			} catch ( IllegalArgumentException e ) {
				responseAccumulator += " unexpected IllegalArgumentException received from getCallback";	
			} catch ( Exception e ) {
				responseAccumulator += " unexpected other exception received from getCallback: " + e.toString();
			} // end try			
			
			// JCA-TA-9048 - getServiceReference returns null when called from a thread not processing an operation of
			//               a service or of a callback
			try {
			    ServiceReference<?> callbackReference = requestContext.getServiceReference();
			    if ( callbackReference != null ) {
			    	responseAccumulator += " unexpected non-null ServiceReference from getServiceReference";
			    } else {
			    	responseAccumulator += " expected null ServiceReference from getServiceReference";
			    }
			} catch ( IllegalArgumentException e ) {
				responseAccumulator += " unexpected IllegalArgumentException received from getServiceReference";	
			} catch ( Exception e ) {
				responseAccumulator += " unexpected other exception received from getServiceReference: " + e.toString();
			} // end try
			
			// Signal completion
			completionLock.lock();
			completionCondition.signalAll();
			completionLock.unlock();
		} // end method run
		
	} // end class separateThreadInvoker

    /**
     * Service operation for Service3WithCallback
     * - used to test API methods called from code processing an operation of a bidirectional service
     */
	public String operation3(String input) {
		String response = "";
		
		// JCA-TA-9041 - getCallbackReference returns ServiceReference typed by the interface of the callback when called from a thread 
		//               processing an operation of a bidirectional service
		try {
		    ServiceReference<?> callbackReference = requestContext.getCallbackReference();
		    if ( callbackReference != null ) {
			    response += " expected non-null ServiceReference from getCallbackReference";
			    if ( callbackReference.getBusinessInterface().equals(Service3Callback.class) ) {
			    	response += " expected ServiceReference has the callback business interface";
			    	// Invoke the callback...
			    	Service3Callback callback = (Service3Callback) callbackReference.getService();
			    	response += callback.callback1("callback");
			    } else {
			    	response += " unexpected ServiceReference does not have the callback business interface";
			    }
		    } else {
			    response += " unexpected null ServiceReference from getCallbackReference";
		    }
		} catch ( IllegalArgumentException e ) {
			response += " unexpected IllegalArgumentException received from getCallbackReference";	
		} catch ( Exception e ) {
			response += " unexpected other exception received from getCallbackReference: " + e.toString();
		} // end try

		// JCA-TA-9044 - getCallback returns a reference proxy typed by the interface of the callback when called from a thread 
		//               processing an operation of a bidirectional service
		try {
			Object callbackReference = requestContext.getCallback();
		    if ( callbackReference != null ) {
			    response += " expected non-null reference proxy from getCallback";
			    if ( callbackReference instanceof Service3Callback ) {
			    	response += " expected reference proxy has the callback business interface";
			    	// Invoke the callback...
			    	Service3Callback callback = (Service3Callback) callbackReference;
			    	response += callback.callback1("callback");
			    } else {
			    	response += " unexpected reference proxy does not have the callback business interface";
			    }
		    } else {
			    response += " unexpected null reference proxy from getCallback";
		    }
		} catch ( IllegalArgumentException e ) {
			response += " unexpected IllegalArgumentException received from getCallback";	
		} catch ( Exception e ) {
			response += " unexpected other exception received from getCallback: " + e.toString();
		} // end try
		
		return response;
	} // end method operation3

	/**
	 * Callback operation for the Service3Callback callback interface
	 * - used to test API calls made from code executing a callback operation
	 * @param input - input String
	 * @return - returned string
	 */
	public String callback1(String input) {
		String response = "";
		
		// JCA-TA-9047 - getServiceReference returns ServiceReference typed by the interface of the callback when called from a thread 
		//               processing an operation of the callback for a bidirectional service
		try {
		    ServiceReference<?> callbackReference = requestContext.getServiceReference();
		    if ( callbackReference != null ) {
			    response += " expected non-null ServiceReference from getServiceReference";
			    Object interfaze = callbackReference.getService();
			    if ( interfaze instanceof Service3Callback ) {
			    	response += " expected ServiceReference has the callback business interface";
			    } else {
			    	response += " unexpected ServiceReference does not have the callback business interface";
			    }
		    } else {
			    response += " unexpected null ServiceReference from getServiceReference";
		    }
		} catch ( IllegalArgumentException e ) {
			response += " unexpected IllegalArgumentException received from getServiceReference";	
		} catch ( Exception e ) {
			response += " unexpected other exception received from getServiceReference: " + e.toString();
		} // end try
		
		return response;
	} // end method callback1

} // end class
