/*
 *
 * Copyright(C) OASIS(R) 2009,2010. All Rights Reserved.
 * OASIS trademark, IPR and other policies apply.    
 *   
 */
package org.oasisopen.sca.test;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import org.oasisopen.sca.annotation.*;
import org.oasisopen.sca.test.Service1;


/**
 * Java component implementation with
 * 1 service with interface Service1,
 * 1 reference (1..1) with interface ParallelService
 * The reference gets called multiple times in parallel when service1 operation1 is invoked and
 * the maximum value of the number of threads simultaneously using the target service is calculated
 */
@Service(Service1.class)
public class parallelCompositeClientImpl implements Service1 {
	
	@Property
	public String serviceName = "service1";
	@Property
	public int invocationCount = 50;		// Number of parallel invocations of the reference reference1 - default = 50
	@Property
	public int maxWaitTime = 10000;		    // Maximum time to wait for parallel invocations to complete - default = 10 seconds
	// Reference with an implied multiplicity of (1..1) since required=true is the default
	@Reference
	public ParallelService reference1 = null;
	
	// A variable which accumulates results - initial value is 0 and gets incremented by
	// each invocation of reference1 that does not return a 0 result
	private AtomicInteger resultAccumulator = new AtomicInteger();
	private AtomicInteger completionCount = new AtomicInteger();
	
	// Lock and associated condition used to signal the original thread that all of the parallel invocations are complete
	private ReentrantLock completionLock = new ReentrantLock();
	private Condition     completionCondition = completionLock.newCondition();
	
	public String operation1(String input) {
		System.out.println("parallel execution with invocation count = " + invocationCount +
		                   " maxWaitTime = " + maxWaitTime );
		ThreadPoolExecutor theExecutor = new ThreadPoolExecutor( invocationCount,
				                                                 invocationCount,
				                                                 maxWaitTime,
				                                                 TimeUnit.MILLISECONDS,
				                                                 new ArrayBlockingQueue<Runnable>( invocationCount ) );
		boolean completed = executeTasks( theExecutor, maxWaitTime );
		if( !completed ) return serviceName + " parallel service invocation failed - tasks not completed";
		
		// The response from the parallel Composite service is a count of the number of threads using the
		// service instance at the point when this invocation returned - need to 
		int result = resultAccumulator.get();
		System.out.println("Final count of simultaneous threads: " + result );
		if( result > 1 ) {
			return serviceName + " COMPOSITE service invocation successfully used by multiple threads simultaneously";
		} else {
			return serviceName + " COMPOSITE service invocation failed - only 1 thread use the instance at a time";
		} // end if
	} // end method operation1	
	
	/**
	 * Execute a set of runnable tasks with the provided Executor, each of which invokes the service targeted by reference1
	 * Wait for the tasks to complete, up to the maximum time indicated by the waitTime parameter
	 * @return true if the tasks all completed successfully, false otherwise (= timeout or interruption) 
	 */
	private boolean executeTasks( ThreadPoolExecutor theExecutor, int waitTime ) {
				
		long startTime = System.currentTimeMillis(); 
		completionLock.lock();
		for( int i=0; i < invocationCount ; i++ ) {
			theExecutor.execute( new parallelServiceInvoker( reference1, i+1 ) );
		} // end for
		System.out.println("Completed creating tasks - took " + ( System.currentTimeMillis() - startTime ) + " ms" );
		
		try {
			// Wait for all the tasks to complete
			completionCondition.await( waitTime, TimeUnit.MILLISECONDS);
			//System.out.println("Count of tasks that returned: " + completionCount.get() );
			//System.out.println("Completed test in " + (System.currentTimeMillis() - startTime ) + " ms" );
			if( completionCount.get() == invocationCount ) return true;
			return false;
		} catch (InterruptedException e) {
			return false;
		} finally {
			completionLock.unlock();
		} // end try

	} // end method allocateTasks()
	
	/**
	 * An inner class which acts as a runnable task for invoking a target service
	 */
	private class parallelServiceInvoker implements Runnable {
		
		ParallelService parallelService;
		int				count;
		
		public parallelServiceInvoker( ParallelService theReference, int theCount ) {
			super();
			parallelService = theReference;
			count = theCount;
			
		} // end constructor

		public void run() {
			// invoke the target parallel service...
			int output = parallelService.operation1("input");
			
			// The response from the parallel Composite service is a count of the number of threads using the
			// service instance at the point when this invocation returned - need to store the maximum value
			// found
			if( output > 0 ) {
				int current;
				// The following stores the new output in the accumulator IF its value is greater that the
				// current value - the loop deals with parallel updates on multiple threads
				do {
					current = resultAccumulator.get();
					if( output <= current ) break;
				} while ( !resultAccumulator.compareAndSet(current, output) );
			} // end if 
			
			int completed = completionCount.incrementAndGet();
			if( completed == invocationCount ) {
				//System.out.println("All tasks completed...");
				completionLock.lock();
				completionCondition.signalAll();
				completionLock.unlock();
			} // end if
			return;
		} // end method run
		
	} // end class parallelServiceInvoker

}
